from airflow.models.baseoperator import BaseOperator
from airflow.hooks.base_hook import BaseHook
from airflow.exceptions import AirflowException
from airflow.utils.decorators import apply_defaults
from airflow.contrib.hooks.bigquery_hook import BigQueryHook

class SQLToSQLOperator(BaseOperator):
    """
    Executes BigQuery SQL queries in a specific BigQuery database

    :param sql: the sql code to be executed (templated)
    :type sql: Can receive a str representing a sql statement,
        a list of str (sql statements), or reference to a template file.
        Template reference are recognized by str ending in '.sql'.
    :param src_conn_id: reference to a source data connection hook.
    :type src_conn_id: str
    :param src_hook_config: a dictionary of properties to apply to the source connection hook. IE {'use_legacy_sql': True}
    :type src_hook_config: dict
    :param cleanup_query: a sql query which will be executed prior to inserting records, this can be used to guarantee idempotency
    :type cleanup_query: str
    :param dest_conn_id: reference to the destination sql database hook. Supported hooks are ['postgres']
    :type dest_conn_id: str
    :param dest_hook_config: a dictionary of properties to apply to the destination connection hook. IE {'use_legacy_sql': False}
    :type dest_hook_config: dict
    :param dest_schema: a db schema name where the destination table resides. Default public
    :type dest_schema: str
    :param dest_table: a table matching the data and schema of the data returned by sql, which the data will be inserted into
    :type dest_table: str
    :param dest_fields: a list of str, naming the fields that the result will be inserted into.
    :type dest_fields: list
    :param batch_size: the number of insert values to batch together to speed up inserts. Default: 1000
    :type batch_size: int

    """

    template_fields = ('sql', 'dest_schema', 'dest_table', 'cleanup_query')
    template_ext = ('.sql', )

    @apply_defaults
    def __init__(self,
                 sql=None,
                 src_conn_id: str = None,
                 src_hook_config: dict = None,
                 dest_conn_id: str = None,
                 dest_hook_config: dict = None,
                 dest_schema: str = 'public',
                 dest_table: str = None,
                 dest_fields: list = None,
                 cleanup_query: str = None,
                 batch_size: int = 1000,
                 *args,
                 **kwargs):
        super(SQLToSQLOperator, self).__init__(*args, **kwargs)
        self.sql = sql
        self.src_conn_id = src_conn_id
        self.src_hook_config = src_hook_config
        self.cleanup_query = cleanup_query
        self.dest_conn_id = dest_conn_id
        self.dest_hook_config = dest_hook_config
        self.dest_schema = dest_schema
        self.dest_table = dest_table
        self.dest_fields = dest_fields
        self.batch_size = batch_size
        self.src_hook = None
        self.dest_hook = None


    def _get_hook(self, connection_id: str, hook_config: dict = None):
        """
        Retrieve the DbAPIHook for the supplied connection_id
        :param connection_id: a string naming the connection id to retrieve
        :param hook_config: a dict of properties that will be applied to the hook
        :return: the DbAPIHook
        """
        conn = BaseHook.get_connection(connection_id)

        allowed_conn_type = {'google_cloud_platform', 'jdbc', 'mssql',
                             'mysql', 'oracle', 'postgres',
                             'sqlite', 'vertica'}
        if conn.conn_type not in allowed_conn_type:
            raise AirflowException("The {} connection type is not supported by SQLToSQLOperator. " +
                                   "Supported connection types: {}".format(connection_id, list(allowed_conn_type)))

        hook = conn.get_hook()

        if hook_config:
            self.log.info('Applying hook properties...')
            for key in hook_config.keys():
                setattr(hook, key, hook_config[key])

            self.log.info('Completed application of hook properties')

        return hook


    def _fetch_from_sql(self):
        """
        Load the source data from the source DbApiHook using the configured SQL
        :return: The result of DbApiHook.get_records
        """
        return self.src_hook.get_records(sql=self.sql)


    def execute(self, context):
        self.log.info('Retrieving the source hook: {}'.format(self.src_conn_id))
        self.src_hook = self._get_hook(connection_id=self.src_conn_id, hook_config=self.src_hook_config)
        self.log.info('Completed retrieval of the source hook')

        self.log.info('Retrieving the destination hook: {}'.format(self.dest_conn_id))
        self.dest_hook = self._get_hook(connection_id=self.dest_conn_id, hook_config=self.dest_hook_config)
        self.log.info('Completed retrieval of the destination hook')

        self.log.info('Reading SQL data: {}'.format(self.sql))
        result = self._fetch_from_sql()
        self.log.info('Completed loading {} records from SQL'.format(len(result)))

        if self.cleanup_query:
            self.log.info('Executing cleanup query: {}'.format(self.cleanup_query))
            self.dest_hook.run(sql=self.cleanup_query)
            self.log.info('Completed execution of cleanup query')

        self.log.info('Inserting retrieved records into destination')        
        self.dest_hook.insert_rows(table='{}.{}'.format(self.dest_schema, self.dest_table),
                                   rows=result,
                                   target_fields=self.dest_fields,
                                   commit_every=self.batch_size
                                    )
        self.log.info('Completed inserting records into destination')




from time import sleep
from datetime import timedelta
from airflow.utils import timezone

from airflow.models import TaskReschedule
from airflow.exceptions import AirflowSensorTimeout, AirflowSkipException, AirflowRescheduleException
from airflow.sensors.sql_sensor import SqlSensor

class SqlSensor(SqlSensor):
    """
    Override of the default SqlSensor with modified 'execute' logic.  Sensor now fails quitely in soft_fail mode
    """
    def execute(self, context):
        started_at = timezone.utcnow()
        if self.reschedule:
            task_reschedules = TaskReschedule.find_for_task_instance(context['ti']) #pragma: no cover
            if task_reschedules:                                                    #pragma: no cover
                started_at = task_reschedules[0].start_date                         #pragma: no cover
        while not self.poke(context):
            if (timezone.utcnow() - started_at).total_seconds() > self.timeout:
                if self.soft_fail and not context['ti'].is_eligible_to_retry():
                    #self._do_skip_downstream_tasks(context)                        # Commented out this line so that skip logic is run as usual
                    self.log.warning("SENSOR TIMED OUT IN SOFT FAIL MODE")
                    raise AirflowSkipException('Snap. Time is OUT.')
                else:
                    raise AirflowSensorTimeout('Snap. Time is OUT.')
            if self.reschedule:                                                     #pragma: no cover
                reschedule_date = timezone.utcnow() + timedelta(                    #pragma: no cover
                    seconds=self.poke_interval)                                     #pragma: no cover
                raise AirflowRescheduleException(reschedule_date)                   #pragma: no cover
            else:
                sleep(self.poke_interval)
        self.log.info("Success criteria met. Exiting.")


import setuptools
import os
from setuptools.command.install import install

if os.environ.get('CI_COMMIT_TAG'):
    version = os.environ['CI_COMMIT_TAG']
else:
    version = os.environ['CI_JOB_ID']

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as req:
    requirements = list(filter(None, [x if 'github.com' not in x else None for x in req.read().split('\n')]))

EXCLUDE_FROM_PACKAGES = ['docs', 'tests*']

EXTRAS_REQUIREMENTS = {
    'default': [
        'apache-airflow==1.10.9'
    ],
    'composer': [
        'apache-airflow==1.10.9-composer'
    ]
}

setuptools.setup(
    name="awehflow",
    version=version.strip('v'),
    author="Philip Perold",
    author_email="philip@spatialedge.co.za",
    description="Configuration based Apache Airflow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(exclude=EXCLUDE_FROM_PACKAGES),
    install_requires=requirements,
    setup_requires=['wheel'],
    extras_require=EXTRAS_REQUIREMENTS,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    include_package_data=True
)



import datetime
import random
import string

import pytz


def merge_dicts(a, b, path=None):
    """merges b into a"""
    if path is None:
        path = []
    for key in b:
        if key in a and (isinstance(a[key], dict) and isinstance(b[key], dict)):
            merge_dicts(a[key], b[key], path + [str(key)])
        else:
            a[key] = b[key]
    return a


def random_string(string_length=10):
    """Generate a random string of fixed length """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for _ in range(string_length))


def uniquify_name(name):
    return "{name}_{random_string}".format(
        name=name,
        random_string=random_string(string_length=6)
    )


def utc_now():
    return pytz.utc.localize(datetime.datetime.utcnow())


class JobStatus:
    RUNNING = 'running'
    FAILURE = 'failure'
    SUCCESS = 'success'



import logging


class EventHandler:
    def handle(self, event):
        try:
            handler_name = event.get('name', '')
            method = getattr(self, handler_name, None)
            if callable(method):
                method(event)
            else:
                self.catch_all(event)
        except Exception as e:
            logging.error('Error handling event: {}'.format(e))

    def catch_all(self, event):
        logging.info('Unhandled <{}> event: {}'.format(self.__class__.__name__, event))
